/**
 * @type {Record<string, string>} yosys 器件映射
 * - key: 器件的 type，比如 $_AND_
 * - value: 器件对应的我们内部的名称，用于检索皮肤系统的
 */
export const DEVICE_MAPPER = {
    "$_AND_": "and",
    "$_XOR_": "xor",
    "$_NOT_": "not",
    "$_OR_": "or",
    "$_NAND_": "nand",
    "$_NOR_": "nor",
    "$_XNOR_": "xnor",
    "$_MUX_": "mux",
    "$_AOI3_": "aoi3",
    "$_OAI3_": "oai3",
    "$_AOI4_": "aoi4",
    "$_OAI4_": "oai4",
    "$_DFF_P_": "dff_pos",
    "$_DFF_N_": "dff_neg",
    "$_SR_NN_": "sr_nn",
    "$_SR_NP_": "sr_np",
    "$_SR_PN_": "sr_pn",
    "$_SR_PP_": "sr_pp"
};

/**
 * @description 模块树对象，直接作为 treeview 的渲染视图加入运算
 * 相比于 YosysNetModule, ModuleTree 只关心和渲染有关的那部分变量，且不可被序列化
 */
export class ModuleTree {
    /**
     * @param {string} name
     * @param {YosysNetModule} rawModule
     */
    constructor(name, rawModule) {
        this.name = name;
        this.rawModule = rawModule;
    }

    get id() {
        return this.name;
    }
}

export class Port {
    /**
     * @description port 的抽象
     * @param {ModuleTree} moduleTree
     * @param {string} name
     * @param {YosysPort} rawPort 
     */
    constructor(moduleTree, name, rawPort) {
        this.moduleTree = moduleTree;
        this.name = name;
        this.rawPort = rawPort;
    }

    get id() {
        return dotConnect(this.moduleTree.id, this.name);
    }
}

export class Cell {
    /**
     * @description 器件的抽象
     * @param {ModuleTree} moduleTree
     * @param {string} name 
     * @param {YosysCell} rawCell 
     */
    constructor(moduleTree, name, rawCell) {
        this.moduleTree = moduleTree;
        this.name = name;
        this.rawCell = rawCell;
    }

    get id() {
        return dotConnect(this.moduleTree.id, this.name);
    }
}

export class Connection {
    /**
     * @description connection 的抽象，一个 connection 代表一个器件的一个端口是如何与外界连接的
     * @param {Cell} cell 
     * @param {string} name 
     * @param {WireId[]} wireIds 
     */
    constructor(cell, name, wireIds) {
        this.cell = cell;
        this.name = name;
        this.wireIds = wireIds;
        
        
    }

    get id() {
        return dotConnect(this.cell.id, this.name);
    }
}

export class Wire {
    /**
     * @description 一根线的抽象
     * @param {Connection} connection 
     * @param {number} index
     * @param {WireId} wireId 
     */
    constructor(connection, index, wireId) {
        this.connection = connection;
        this.index = index;
        this.wireId = wireId;
    }

    get id() {
        if (typeof this.wireId === 'string') {
            // 常数连接
            const num = parseInt(this.wireId);
            return dotConnect(this.connection.id, this.index, 'c-' + num);
        } else {
            // 普通连接
            return dotConnect(this.connection.id, this.wireId);
        }
    }
}

/**
 * @description 使用 . 把各个部分连接起来
 * @param {any[]} args
 */
export function dotConnect(...args) {
    const stringArgs = args.map(arg => arg.toString());
    return stringArgs.join('.');
}